# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io` -- Input/output functions for interaction with user's script
=================================================================================

.. module:: io
   :platform: Unix
   :synopsis: Parsers and formatters to be used with user's input files or cli
      arguments and expected outputs.
"""
