# -*- coding: utf-8 -*-
"""
:mod:`orion.core.evc` -- Experiment Version Control System
==========================================================

.. module:: evc
   :platform: Unix
   :synopsis: Organize related experiments inside a tree-structure using adapters to maintain
      compatibility between them.
"""
