#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.db` -- Functions that define console scripts for database
==============================================================================

.. module:: cli
   :platform: Unix
   :synopsis: Commands to maintain database

"""
